@@ -7,7 +7,7 @@ import pandas as pd
 from .coding import strings, times, variables
 from .coding.variables import SerializationWarning, pop_to
 from .core import duck_array_ops, indexing
-from .core.common import contains_cftime_datetimes
+from .core.common import _contains_datetime_like_objects, contains_cftime_datetimes
 from .core.pycompat import is_duck_dask_array
 from .core.variable import IndexVariable, Variable, as_variable
 
@@ -157,8 +157,12 @@ def _infer_dtype(array, name=None):
         return np.dtype(float)
 
     element = array[(0,) * array.ndim]
-    if isinstance(element, (bytes, str)):
-        return strings.create_vlen_dtype(type(element))
+    # We use the base types to avoid subclasses of bytes and str (which might
+    # not play nice with e.g. hdf5 datatypes), such as those from numpy
+    if isinstance(element, bytes):
+        return strings.create_vlen_dtype(bytes)
+    elif isinstance(element, str):
+        return strings.create_vlen_dtype(str)
 
     dtype = np.array(element).dtype
     if dtype.kind != "O":
@@ -336,6 +340,11 @@ def decode_cf_variable(
         A variable holding the decoded equivalent of var.
     """
     var = as_variable(var)
+
+    # Ensure datetime-like Variables are passed through unmodified (GH 6453)
+    if _contains_datetime_like_objects(var):
+        return var
+
     original_dtype = var.dtype
 
     if decode_timedelta is None:
@@ -766,7 +775,7 @@ def _encode_coordinates(variables, attributes, non_dim_coord_names):
         # this will copy coordinates from encoding to attrs if "coordinates" in attrs
         # after the next line, "coordinates" is never in encoding
         # we get support for attrs["coordinates"] for free.
-        coords_str = pop_to(encoding, attrs, "coordinates")
+        coords_str = pop_to(encoding, attrs, "coordinates") or attrs.get("coordinates")
         if not coords_str and variable_coordinates[name]:
             coordinates_text = " ".join(
                 str(coord_name)
