@@ -13,6 +13,7 @@ from typing import (
     Iterator,
     Mapping,
     TypeVar,
+    Union,
     overload,
 )
 
@@ -20,10 +21,9 @@ import numpy as np
 import pandas as pd
 
 from . import dtypes, duck_array_ops, formatting, formatting_html, ops
-from .npcompat import DTypeLike
+from .npcompat import DTypeLike, DTypeLikeSave
 from .options import OPTIONS, _get_keep_attrs
 from .pycompat import is_duck_dask_array
-from .rolling_exp import RollingExp
 from .utils import Frozen, either_dict_or_kwargs, is_scalar
 
 try:
@@ -36,13 +36,18 @@ ALL_DIMS = ...
 
 
 if TYPE_CHECKING:
+    import datetime
+
     from .dataarray import DataArray
     from .dataset import Dataset
-    from .types import T_DataWithCoords, T_Xarray
+    from .indexes import Index
+    from .resample import Resample
+    from .rolling_exp import RollingExp
+    from .types import ScalarOrArray, SideOptions, T_DataWithCoords
     from .variable import Variable
-    from .weighted import Weighted
 
 
+T_Resample = TypeVar("T_Resample", bound="Resample")
 C = TypeVar("C")
 T = TypeVar("T")
 
@@ -55,12 +60,14 @@ class ImplementsArrayReduce:
         if include_skipna:
 
             def wrapped_func(self, dim=None, axis=None, skipna=None, **kwargs):
-                return self.reduce(func, dim, axis, skipna=skipna, **kwargs)
+                return self.reduce(
+                    func=func, dim=dim, axis=axis, skipna=skipna, **kwargs
+                )
 
         else:
 
             def wrapped_func(self, dim=None, axis=None, **kwargs):  # type: ignore[misc]
-                return self.reduce(func, dim, axis, **kwargs)
+                return self.reduce(func=func, dim=dim, axis=axis, **kwargs)
 
         return wrapped_func
 
@@ -93,13 +100,19 @@ class ImplementsDatasetReduce:
 
             def wrapped_func(self, dim=None, skipna=None, **kwargs):
                 return self.reduce(
-                    func, dim, skipna=skipna, numeric_only=numeric_only, **kwargs
+                    func=func,
+                    dim=dim,
+                    skipna=skipna,
+                    numeric_only=numeric_only,
+                    **kwargs,
                 )
 
         else:
 
             def wrapped_func(self, dim=None, **kwargs):  # type: ignore[misc]
-                return self.reduce(func, dim, numeric_only=numeric_only, **kwargs)
+                return self.reduce(
+                    func=func, dim=dim, numeric_only=numeric_only, **kwargs
+                )
 
         return wrapped_func
 
@@ -150,6 +163,10 @@ class AbstractArray:
             return f"<pre>{escape(repr(self))}</pre>"
         return formatting_html.array_repr(self)
 
+    def __format__(self: Any, format_spec: str) -> str:
+        # we use numpy: scalars will print fine and arrays will raise
+        return self.values.__format__(format_spec)
+
     def _iter(self: Any) -> Iterator[Any]:
         for n in range(len(self)):
             yield self[n]
@@ -184,7 +201,7 @@ class AbstractArray:
             raise ValueError(f"{dim!r} not found in array dimensions {self.dims!r}")
 
     @property
-    def sizes(self: Any) -> Mapping[Hashable, int]:
+    def sizes(self: Any) -> Frozen[Hashable, int]:
         """Ordered mapping from dimension names to lengths.
 
         Immutable.
@@ -340,15 +357,16 @@ class DataWithCoords(AttrAccessMixin):
     """Shared base class for Dataset and DataArray."""
 
     _close: Callable[[], None] | None
+    _indexes: dict[Hashable, Index]
 
     __slots__ = ("_close",)
 
     def squeeze(
-        self,
+        self: T_DataWithCoords,
         dim: Hashable | Iterable[Hashable] | None = None,
         drop: bool = False,
         axis: int | Iterable[int] | None = None,
-    ):
+    ) -> T_DataWithCoords:
         """Return a new object with squeezed data.
 
         Parameters
@@ -357,7 +375,7 @@ class DataWithCoords(AttrAccessMixin):
             Selects a subset of the length one dimensions. If a dimension is
             selected with length greater than one, an error is raised. If
             None, all length one dimensions are squeezed.
-        drop : bool, optional
+        drop : bool, default: False
             If ``drop=True``, drop squeezed coordinates instead of making them
             scalar.
         axis : None or int or iterable of int, optional
@@ -376,12 +394,33 @@ class DataWithCoords(AttrAccessMixin):
         dims = get_squeeze_dims(self, dim, axis)
         return self.isel(drop=drop, **{d: 0 for d in dims})
 
-    def clip(self, min=None, max=None, *, keep_attrs: bool = None):
+    def clip(
+        self: T_DataWithCoords,
+        min: ScalarOrArray | None = None,
+        max: ScalarOrArray | None = None,
+        *,
+        keep_attrs: bool | None = None,
+    ) -> T_DataWithCoords:
         """
         Return an array whose values are limited to ``[min, max]``.
         At least one of max or min must be given.
 
-        Refer to `numpy.clip` for full documentation.
+        Parameters
+        ----------
+        min : None or Hashable, optional
+            Minimum value. If None, no lower clipping is performed.
+        max : None or Hashable, optional
+            Maximum value. If None, no upper clipping is performed.
+        keep_attrs : bool or None, optional
+            If True, the attributes (`attrs`) will be copied from
+            the original object to the new one. If False, the new
+            object will be returned without attributes.
+
+        Returns
+        -------
+        clipped : same type as caller
+            This object, but with with values < min are replaced with min,
+            and those > max with max.
 
         See Also
         --------
@@ -404,7 +443,7 @@ class DataWithCoords(AttrAccessMixin):
             raise KeyError(key)
 
         try:
-            return self.xindexes[key].to_pandas_index()
+            return self._indexes[key].to_pandas_index()
         except KeyError:
             return pd.Index(range(self.sizes[key]), name=key)
 
@@ -413,7 +452,11 @@ class DataWithCoords(AttrAccessMixin):
     ) -> dict[Hashable, T]:
         return {k: v(self) if callable(v) else v for k, v in kwargs.items()}
 
-    def assign_coords(self, coords=None, **coords_kwargs):
+    def assign_coords(
+        self: T_DataWithCoords,
+        coords: Mapping[Any, Any] | None = None,
+        **coords_kwargs: Any,
+    ) -> T_DataWithCoords:
         """Assign new coordinates to this object.
 
         Returns a new object with all the original data in addition to the new
@@ -421,7 +464,7 @@ class DataWithCoords(AttrAccessMixin):
 
         Parameters
         ----------
-        coords : dict, optional
+        coords : dict-like or None, optional
             A dict where the keys are the names of the coordinates
             with the new values to assign. If the values are callable, they are
             computed on this object and assigned to new coordinate variables.
@@ -442,7 +485,7 @@ class DataWithCoords(AttrAccessMixin):
 
         Examples
         --------
-        Convert longitude coordinates from 0-359 to -180-179:
+        Convert `DataArray` longitude coordinates from 0-359 to -180-179:
 
         >>> da = xr.DataArray(
         ...     np.random.rand(4),
@@ -482,6 +525,54 @@ class DataWithCoords(AttrAccessMixin):
 
         >>> _ = da.assign_coords({"lon_2": ("lon", lon_2)})
 
+        Note the same method applies to `Dataset` objects.
+
+        Convert `Dataset` longitude coordinates from 0-359 to -180-179:
+
+        >>> temperature = np.linspace(20, 32, num=16).reshape(2, 2, 4)
+        >>> precipitation = 2 * np.identity(4).reshape(2, 2, 4)
+        >>> ds = xr.Dataset(
+        ...     data_vars=dict(
+        ...         temperature=(["x", "y", "time"], temperature),
+        ...         precipitation=(["x", "y", "time"], precipitation),
+        ...     ),
+        ...     coords=dict(
+        ...         lon=(["x", "y"], [[260.17, 260.68], [260.21, 260.77]]),
+        ...         lat=(["x", "y"], [[42.25, 42.21], [42.63, 42.59]]),
+        ...         time=pd.date_range("2014-09-06", periods=4),
+        ...         reference_time=pd.Timestamp("2014-09-05"),
+        ...     ),
+        ...     attrs=dict(description="Weather-related data"),
+        ... )
+        >>> ds
+        <xarray.Dataset>
+        Dimensions:         (x: 2, y: 2, time: 4)
+        Coordinates:
+            lon             (x, y) float64 260.2 260.7 260.2 260.8
+            lat             (x, y) float64 42.25 42.21 42.63 42.59
+          * time            (time) datetime64[ns] 2014-09-06 2014-09-07 ... 2014-09-09
+            reference_time  datetime64[ns] 2014-09-05
+        Dimensions without coordinates: x, y
+        Data variables:
+            temperature     (x, y, time) float64 20.0 20.8 21.6 22.4 ... 30.4 31.2 32.0
+            precipitation   (x, y, time) float64 2.0 0.0 0.0 0.0 0.0 ... 0.0 0.0 0.0 2.0
+        Attributes:
+            description:  Weather-related data
+        >>> ds.assign_coords(lon=(((ds.lon + 180) % 360) - 180))
+        <xarray.Dataset>
+        Dimensions:         (x: 2, y: 2, time: 4)
+        Coordinates:
+            lon             (x, y) float64 -99.83 -99.32 -99.79 -99.23
+            lat             (x, y) float64 42.25 42.21 42.63 42.59
+          * time            (time) datetime64[ns] 2014-09-06 2014-09-07 ... 2014-09-09
+            reference_time  datetime64[ns] 2014-09-05
+        Dimensions without coordinates: x, y
+        Data variables:
+            temperature     (x, y, time) float64 20.0 20.8 21.6 22.4 ... 30.4 31.2 32.0
+            precipitation   (x, y, time) float64 2.0 0.0 0.0 0.0 0.0 ... 0.0 0.0 0.0 2.0
+        Attributes:
+            description:  Weather-related data
+
         Notes
         -----
         Since ``coords_kwargs`` is a dictionary, the order of your arguments
@@ -495,13 +586,15 @@ class DataWithCoords(AttrAccessMixin):
         Dataset.assign
         Dataset.swap_dims
         """
-        coords_kwargs = either_dict_or_kwargs(coords, coords_kwargs, "assign_coords")
+        coords_combined = either_dict_or_kwargs(coords, coords_kwargs, "assign_coords")
         data = self.copy(deep=False)
-        results = self._calc_assign_results(coords_kwargs)
+        results: dict[Hashable, Any] = self._calc_assign_results(coords_combined)
         data.coords.update(results)
         return data
 
-    def assign_attrs(self, *args, **kwargs):
+    def assign_attrs(
+        self: T_DataWithCoords, *args: Any, **kwargs: Any
+    ) -> T_DataWithCoords:
         """Assign new attrs to this object.
 
         Returns a new object equivalent to ``self.attrs.update(*args, **kwargs)``.
@@ -529,8 +622,8 @@ class DataWithCoords(AttrAccessMixin):
     def pipe(
         self,
         func: Callable[..., T] | tuple[Callable[..., T], str],
-        *args,
-        **kwargs,
+        *args: Any,
+        **kwargs: Any,
     ) -> T:
         """
         Apply ``func(self, *args, **kwargs)``
@@ -658,242 +751,12 @@ class DataWithCoords(AttrAccessMixin):
         else:
             return func(self, *args, **kwargs)
 
-    def groupby(self, group, squeeze: bool = True, restore_coord_dims: bool = None):
-        """Returns a GroupBy object for performing grouped operations.
-
-        Parameters
-        ----------
-        group : str, DataArray or IndexVariable
-            Array whose unique values should be used to group this array. If a
-            string, must be the name of a variable contained in this dataset.
-        squeeze : bool, optional
-            If "group" is a dimension of any arrays in this dataset, `squeeze`
-            controls whether the subarrays have a dimension of length 1 along
-            that dimension or if the dimension is squeezed out.
-        restore_coord_dims : bool, optional
-            If True, also restore the dimension order of multi-dimensional
-            coordinates.
-
-        Returns
-        -------
-        grouped
-            A `GroupBy` object patterned after `pandas.GroupBy` that can be
-            iterated over in the form of `(unique_value, grouped_array)` pairs.
-
-        Examples
-        --------
-        Calculate daily anomalies for daily data:
-
-        >>> da = xr.DataArray(
-        ...     np.linspace(0, 1826, num=1827),
-        ...     coords=[pd.date_range("1/1/2000", "31/12/2004", freq="D")],
-        ...     dims="time",
-        ... )
-        >>> da
-        <xarray.DataArray (time: 1827)>
-        array([0.000e+00, 1.000e+00, 2.000e+00, ..., 1.824e+03, 1.825e+03,
-               1.826e+03])
-        Coordinates:
-          * time     (time) datetime64[ns] 2000-01-01 2000-01-02 ... 2004-12-31
-        >>> da.groupby("time.dayofyear") - da.groupby("time.dayofyear").mean("time")
-        <xarray.DataArray (time: 1827)>
-        array([-730.8, -730.8, -730.8, ...,  730.2,  730.2,  730.5])
-        Coordinates:
-          * time       (time) datetime64[ns] 2000-01-01 2000-01-02 ... 2004-12-31
-            dayofyear  (time) int64 1 2 3 4 5 6 7 8 ... 359 360 361 362 363 364 365 366
-
-        See Also
-        --------
-        core.groupby.DataArrayGroupBy
-        core.groupby.DatasetGroupBy
-        """
-        # While we don't generally check the type of every arg, passing
-        # multiple dimensions as multiple arguments is common enough, and the
-        # consequences hidden enough (strings evaluate as true) to warrant
-        # checking here.
-        # A future version could make squeeze kwarg only, but would face
-        # backward-compat issues.
-        if not isinstance(squeeze, bool):
-            raise TypeError(
-                f"`squeeze` must be True or False, but {squeeze} was supplied"
-            )
-
-        return self._groupby_cls(
-            self, group, squeeze=squeeze, restore_coord_dims=restore_coord_dims
-        )
-
-    def groupby_bins(
-        self,
-        group,
-        bins,
-        right: bool = True,
-        labels=None,
-        precision: int = 3,
-        include_lowest: bool = False,
-        squeeze: bool = True,
-        restore_coord_dims: bool = None,
-    ):
-        """Returns a GroupBy object for performing grouped operations.
-
-        Rather than using all unique values of `group`, the values are discretized
-        first by applying `pandas.cut` [1]_ to `group`.
-
-        Parameters
-        ----------
-        group : str, DataArray or IndexVariable
-            Array whose binned values should be used to group this array. If a
-            string, must be the name of a variable contained in this dataset.
-        bins : int or array-like
-            If bins is an int, it defines the number of equal-width bins in the
-            range of x. However, in this case, the range of x is extended by .1%
-            on each side to include the min or max values of x. If bins is a
-            sequence it defines the bin edges allowing for non-uniform bin
-            width. No extension of the range of x is done in this case.
-        right : bool, default: True
-            Indicates whether the bins include the rightmost edge or not. If
-            right == True (the default), then the bins [1,2,3,4] indicate
-            (1,2], (2,3], (3,4].
-        labels : array-like or bool, default: None
-            Used as labels for the resulting bins. Must be of the same length as
-            the resulting bins. If False, string bin labels are assigned by
-            `pandas.cut`.
-        precision : int
-            The precision at which to store and display the bins labels.
-        include_lowest : bool
-            Whether the first interval should be left-inclusive or not.
-        squeeze : bool, default: True
-            If "group" is a dimension of any arrays in this dataset, `squeeze`
-            controls whether the subarrays have a dimension of length 1 along
-            that dimension or if the dimension is squeezed out.
-        restore_coord_dims : bool, optional
-            If True, also restore the dimension order of multi-dimensional
-            coordinates.
-
-        Returns
-        -------
-        grouped
-            A `GroupBy` object patterned after `pandas.GroupBy` that can be
-            iterated over in the form of `(unique_value, grouped_array)` pairs.
-            The name of the group has the added suffix `_bins` in order to
-            distinguish it from the original variable.
-
-        References
-        ----------
-        .. [1] http://pandas.pydata.org/pandas-docs/stable/generated/pandas.cut.html
-        """
-        return self._groupby_cls(
-            self,
-            group,
-            squeeze=squeeze,
-            bins=bins,
-            restore_coord_dims=restore_coord_dims,
-            cut_kwargs={
-                "right": right,
-                "labels": labels,
-                "precision": precision,
-                "include_lowest": include_lowest,
-            },
-        )
-
-    def weighted(self: T_DataWithCoords, weights: DataArray) -> Weighted[T_Xarray]:
-        """
-        Weighted operations.
-
-        Parameters
-        ----------
-        weights : DataArray
-            An array of weights associated with the values in this Dataset.
-            Each value in the data contributes to the reduction operation
-            according to its associated weight.
-
-        Notes
-        -----
-        ``weights`` must be a DataArray and cannot contain missing values.
-        Missing values can be replaced by ``weights.fillna(0)``.
-        """
-
-        return self._weighted_cls(self, weights)
-
-    def rolling(
-        self,
-        dim: Mapping[Any, int] = None,
-        min_periods: int = None,
-        center: bool | Mapping[Any, bool] = False,
-        **window_kwargs: int,
-    ):
-        """
-        Rolling window object.
-
-        Parameters
-        ----------
-        dim : dict, optional
-            Mapping from the dimension name to create the rolling iterator
-            along (e.g. `time`) to its moving window size.
-        min_periods : int, default: None
-            Minimum number of observations in window required to have a value
-            (otherwise result is NA). The default, None, is equivalent to
-            setting min_periods equal to the size of the window.
-        center : bool or mapping, default: False
-            Set the labels at the center of the window.
-        **window_kwargs : optional
-            The keyword arguments form of ``dim``.
-            One of dim or window_kwargs must be provided.
-
-        Returns
-        -------
-        core.rolling.DataArrayRolling or core.rolling.DatasetRolling
-            A rolling object (``DataArrayRolling`` for ``DataArray``,
-            ``DatasetRolling`` for ``Dataset``)
-
-        Examples
-        --------
-        Create rolling seasonal average of monthly data e.g. DJF, JFM, ..., SON:
-
-        >>> da = xr.DataArray(
-        ...     np.linspace(0, 11, num=12),
-        ...     coords=[
-        ...         pd.date_range(
-        ...             "1999-12-15",
-        ...             periods=12,
-        ...             freq=pd.DateOffset(months=1),
-        ...         )
-        ...     ],
-        ...     dims="time",
-        ... )
-        >>> da
-        <xarray.DataArray (time: 12)>
-        array([ 0.,  1.,  2.,  3.,  4.,  5.,  6.,  7.,  8.,  9., 10., 11.])
-        Coordinates:
-          * time     (time) datetime64[ns] 1999-12-15 2000-01-15 ... 2000-11-15
-        >>> da.rolling(time=3, center=True).mean()
-        <xarray.DataArray (time: 12)>
-        array([nan,  1.,  2.,  3.,  4.,  5.,  6.,  7.,  8.,  9., 10., nan])
-        Coordinates:
-          * time     (time) datetime64[ns] 1999-12-15 2000-01-15 ... 2000-11-15
-
-        Remove the NaNs using ``dropna()``:
-
-        >>> da.rolling(time=3, center=True).mean().dropna("time")
-        <xarray.DataArray (time: 10)>
-        array([ 1.,  2.,  3.,  4.,  5.,  6.,  7.,  8.,  9., 10.])
-        Coordinates:
-          * time     (time) datetime64[ns] 2000-01-15 2000-02-15 ... 2000-10-15
-
-        See Also
-        --------
-        core.rolling.DataArrayRolling
-        core.rolling.DatasetRolling
-        """
-
-        dim = either_dict_or_kwargs(dim, window_kwargs, "rolling")
-        return self._rolling_cls(self, dim, min_periods=min_periods, center=center)
-
     def rolling_exp(
-        self,
+        self: T_DataWithCoords,
         window: Mapping[Any, int] = None,
         window_type: str = "span",
         **window_kwargs,
-    ):
+    ) -> RollingExp[T_DataWithCoords]:
         """
         Exponentially-weighted moving window.
         Similar to EWM in pandas
@@ -917,6 +780,7 @@ class DataWithCoords(AttrAccessMixin):
         --------
         core.rolling_exp.RollingExp
         """
+        from . import rolling_exp
 
         if "keep_attrs" in window_kwargs:
             warnings.warn(
@@ -927,96 +791,21 @@ class DataWithCoords(AttrAccessMixin):
 
         window = either_dict_or_kwargs(window, window_kwargs, "rolling_exp")
 
-        return RollingExp(self, window, window_type)
+        return rolling_exp.RollingExp(self, window, window_type)
 
-    def coarsen(
+    def _resample(
         self,
-        dim: Mapping[Any, int] = None,
-        boundary: str = "exact",
-        side: str | Mapping[Any, str] = "left",
-        coord_func: str = "mean",
-        **window_kwargs: int,
-    ):
-        """
-        Coarsen object.
-
-        Parameters
-        ----------
-        dim : mapping of hashable to int, optional
-            Mapping from the dimension name to the window size.
-        boundary : {"exact", "trim", "pad"}, default: "exact"
-            If 'exact', a ValueError will be raised if dimension size is not a
-            multiple of the window size. If 'trim', the excess entries are
-            dropped. If 'pad', NA will be padded.
-        side : {"left", "right"} or mapping of str to {"left", "right"}
-        coord_func : str or mapping of hashable to str, default: "mean"
-            function (name) that is applied to the coordinates,
-            or a mapping from coordinate name to function (name).
-
-        Returns
-        -------
-        core.rolling.DataArrayCoarsen or core.rolling.DatasetCoarsen
-            A coarsen object (``DataArrayCoarsen`` for ``DataArray``,
-            ``DatasetCoarsen`` for ``Dataset``)
-
-        Examples
-        --------
-        Coarsen the long time series by averaging over every four days.
-
-        >>> da = xr.DataArray(
-        ...     np.linspace(0, 364, num=364),
-        ...     dims="time",
-        ...     coords={"time": pd.date_range("1999-12-15", periods=364)},
-        ... )
-        >>> da  # +doctest: ELLIPSIS
-        <xarray.DataArray (time: 364)>
-        array([  0.        ,   1.00275482,   2.00550964,   3.00826446,
-                 4.01101928,   5.0137741 ,   6.01652893,   7.01928375,
-                 8.02203857,   9.02479339,  10.02754821,  11.03030303,
-        ...
-               356.98071625, 357.98347107, 358.9862259 , 359.98898072,
-               360.99173554, 361.99449036, 362.99724518, 364.        ])
-        Coordinates:
-          * time     (time) datetime64[ns] 1999-12-15 1999-12-16 ... 2000-12-12
-        >>> da.coarsen(time=3, boundary="trim").mean()  # +doctest: ELLIPSIS
-        <xarray.DataArray (time: 121)>
-        array([  1.00275482,   4.01101928,   7.01928375,  10.02754821,
-                13.03581267,  16.04407713,  19.0523416 ,  22.06060606,
-                25.06887052,  28.07713499,  31.08539945,  34.09366391,
-        ...
-               349.96143251, 352.96969697, 355.97796143, 358.9862259 ,
-               361.99449036])
-        Coordinates:
-          * time     (time) datetime64[ns] 1999-12-16 1999-12-19 ... 2000-12-10
-        >>>
-
-        See Also
-        --------
-        core.rolling.DataArrayCoarsen
-        core.rolling.DatasetCoarsen
-        """
-
-        dim = either_dict_or_kwargs(dim, window_kwargs, "coarsen")
-        return self._coarsen_cls(
-            self,
-            dim,
-            boundary=boundary,
-            side=side,
-            coord_func=coord_func,
-        )
-
-    def resample(
-        self,
-        indexer: Mapping[Any, str] = None,
-        skipna=None,
-        closed: str = None,
-        label: str = None,
-        base: int = 0,
-        keep_attrs: bool = None,
-        loffset=None,
-        restore_coord_dims: bool = None,
+        resample_cls: type[T_Resample],
+        indexer: Mapping[Any, str] | None,
+        skipna: bool | None,
+        closed: SideOptions | None,
+        label: SideOptions | None,
+        base: int,
+        keep_attrs: bool | None,
+        loffset: datetime.timedelta | str | None,
+        restore_coord_dims: bool | None,
         **indexer_kwargs: str,
-    ):
+    ) -> T_Resample:
         """Returns a Resample object for performing resampling operations.
 
         Handles both downsampling and upsampling. The resampled
@@ -1108,7 +897,7 @@ class DataWithCoords(AttrAccessMixin):
 
         References
         ----------
-        .. [1] http://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases
+        .. [1] https://pandas.pydata.org/docs/user_guide/timeseries.html#dateoffset-objects
         """
         # TODO support non-string indexer after removing the old API.
 
@@ -1140,7 +929,7 @@ class DataWithCoords(AttrAccessMixin):
             raise ValueError("Resampling only supported along single dimensions.")
         dim, freq = next(iter(indexer.items()))
 
-        dim_name = dim
+        dim_name: Hashable = dim
         dim_coord = self[dim]
 
         # TODO: remove once pandas=1.1 is the minimum required version
@@ -1151,8 +940,7 @@ class DataWithCoords(AttrAccessMixin):
                 category=FutureWarning,
             )
 
-            # TODO (benbovy - flexible indexes): update when CFTimeIndex is an xarray Index subclass
-            if isinstance(self.xindexes[dim_name].to_pandas_index(), CFTimeIndex):
+            if isinstance(self._indexes[dim_name].to_pandas_index(), CFTimeIndex):
                 from .resample_cftime import CFTimeGrouper
 
                 grouper = CFTimeGrouper(freq, closed, label, base, loffset)
@@ -1163,7 +951,7 @@ class DataWithCoords(AttrAccessMixin):
         group = DataArray(
             dim_coord, coords=dim_coord.coords, dims=dim_coord.dims, name=RESAMPLE_DIM
         )
-        resampler = self._resample_cls(
+        return resample_cls(
             self,
             group=group,
             dim=dim_name,
@@ -1172,9 +960,9 @@ class DataWithCoords(AttrAccessMixin):
             restore_coord_dims=restore_coord_dims,
         )
 
-        return resampler
-
-    def where(self, cond, other=dtypes.NA, drop: bool = False):
+    def where(
+        self: T_DataWithCoords, cond: Any, other: Any = dtypes.NA, drop: bool = False
+    ) -> T_DataWithCoords:
         """Filter elements from this object according to a condition.
 
         This operation follows the normal broadcasting and alignment rules that
@@ -1188,10 +976,9 @@ class DataWithCoords(AttrAccessMixin):
         other : scalar, DataArray or Dataset, optional
             Value to use for locations in this object where ``cond`` is False.
             By default, these locations filled with NA.
-        drop : bool, optional
+        drop : bool, default: False
             If True, coordinate labels that only correspond to False values of
-            the condition are dropped from the result. Mutually exclusive with
-            ``other``.
+            the condition are dropped from the result.
 
         Returns
         -------
@@ -1244,6 +1031,14 @@ class DataWithCoords(AttrAccessMixin):
                [15., nan, nan, nan]])
         Dimensions without coordinates: x, y
 
+        >>> a.where(a.x + a.y < 4, -1, drop=True)
+        <xarray.DataArray (x: 4, y: 4)>
+        array([[ 0,  1,  2,  3],
+               [ 5,  6,  7, -1],
+               [10, 11, -1, -1],
+               [15, -1, -1, -1]])
+        Dimensions without coordinates: x, y
+
         See Also
         --------
         numpy.where : corresponding numpy function
@@ -1257,26 +1052,28 @@ class DataWithCoords(AttrAccessMixin):
             cond = cond(self)
 
         if drop:
-            if other is not dtypes.NA:
-                raise ValueError("cannot set `other` if drop=True")
-
             if not isinstance(cond, (Dataset, DataArray)):
                 raise TypeError(
                     f"cond argument is {cond!r} but must be a {Dataset!r} or {DataArray!r}"
                 )
 
-            # align so we can use integer indexing
-            self, cond = align(self, cond)
+            self, cond = align(self, cond)  # type: ignore[assignment]
 
-            # get cond with the minimal size needed for the Dataset
-            if isinstance(cond, Dataset):
-                clipcond = cond.to_array().any("variable")
-            else:
-                clipcond = cond
+            def _dataarray_indexer(dim: Hashable) -> DataArray:
+                return cond.any(dim=(d for d in cond.dims if d != dim))
 
-            # clip the data corresponding to coordinate dims that are not used
-            nonzeros = zip(clipcond.dims, np.nonzero(clipcond.values))
-            indexers = {k: np.unique(v) for k, v in nonzeros}
+            def _dataset_indexer(dim: Hashable) -> DataArray:
+                cond_wdim = cond.drop(var for var in cond if dim not in cond[var].dims)
+                keepany = cond_wdim.any(dim=(d for d in cond.dims.keys() if d != dim))
+                return keepany.to_array().any("variable")
+
+            _get_indexer = (
+                _dataarray_indexer if isinstance(cond, DataArray) else _dataset_indexer
+            )
+
+            indexers = {}
+            for dim in cond.sizes.keys():
+                indexers[dim] = _get_indexer(dim)
 
             self = self.isel(**indexers)
             cond = cond.isel(**indexers)
@@ -1299,15 +1096,24 @@ class DataWithCoords(AttrAccessMixin):
         """
         self._close = close
 
-    def close(self: Any) -> None:
+    def close(self) -> None:
         """Release any resources linked to this object."""
         if self._close is not None:
             self._close()
         self._close = None
 
-    def isnull(self, keep_attrs: bool = None):
+    def isnull(
+        self: T_DataWithCoords, keep_attrs: bool | None = None
+    ) -> T_DataWithCoords:
         """Test each value in the array for whether it is a missing value.
 
+        Parameters
+        ----------
+        keep_attrs : bool or None, optional
+            If True, the attributes (`attrs`) will be copied from
+            the original object to the new one. If False, the new
+            object will be returned without attributes.
+
         Returns
         -------
         isnull : DataArray or Dataset
@@ -1341,9 +1147,18 @@ class DataWithCoords(AttrAccessMixin):
             keep_attrs=keep_attrs,
         )
 
-    def notnull(self, keep_attrs: bool = None):
+    def notnull(
+        self: T_DataWithCoords, keep_attrs: bool | None = None
+    ) -> T_DataWithCoords:
         """Test each value in the array for whether it is not a missing value.
 
+        Parameters
+        ----------
+        keep_attrs : bool or None, optional
+            If True, the attributes (`attrs`) will be copied from
+            the original object to the new one. If False, the new
+            object will be returned without attributes.
+
         Returns
         -------
         notnull : DataArray or Dataset
@@ -1377,7 +1192,7 @@ class DataWithCoords(AttrAccessMixin):
             keep_attrs=keep_attrs,
         )
 
-    def isin(self, test_elements):
+    def isin(self: T_DataWithCoords, test_elements: Any) -> T_DataWithCoords:
         """Tests each value in the array for whether it is in test elements.
 
         Parameters
@@ -1428,7 +1243,7 @@ class DataWithCoords(AttrAccessMixin):
         )
 
     def astype(
-        self: T,
+        self: T_DataWithCoords,
         dtype,
         *,
         order=None,
@@ -1436,7 +1251,7 @@ class DataWithCoords(AttrAccessMixin):
         subok=None,
         copy=None,
         keep_attrs=True,
-    ) -> T:
+    ) -> T_DataWithCoords:
         """
         Copy of the xarray object, with data cast to a specified type.
         Leaves coordinate dtype unchanged.
@@ -1503,7 +1318,7 @@ class DataWithCoords(AttrAccessMixin):
             dask="allowed",
         )
 
-    def __enter__(self: T) -> T:
+    def __enter__(self: T_DataWithCoords) -> T_DataWithCoords:
         return self
 
     def __exit__(self, exc_type, exc_value, traceback) -> None:
@@ -1514,26 +1329,51 @@ class DataWithCoords(AttrAccessMixin):
         raise NotImplementedError()
 
 
+DTypeMaybeMapping = Union[DTypeLikeSave, Mapping[Any, DTypeLikeSave]]
+
+
 @overload
 def full_like(
-    other: Dataset,
-    fill_value,
-    dtype: DTypeLike | Mapping[Any, DTypeLike] = None,
+    other: DataArray, fill_value: Any, dtype: DTypeLikeSave = None
+) -> DataArray:
+    ...
+
+
+@overload
+def full_like(
+    other: Dataset, fill_value: Any, dtype: DTypeMaybeMapping = None
 ) -> Dataset:
     ...
 
 
 @overload
-def full_like(other: DataArray, fill_value, dtype: DTypeLike = None) -> DataArray:
+def full_like(
+    other: Variable, fill_value: Any, dtype: DTypeLikeSave = None
+) -> Variable:
     ...
 
 
 @overload
-def full_like(other: Variable, fill_value, dtype: DTypeLike = None) -> Variable:
+def full_like(
+    other: Dataset | DataArray, fill_value: Any, dtype: DTypeMaybeMapping = None
+) -> Dataset | DataArray:
     ...
 
 
-def full_like(other, fill_value, dtype=None):
+@overload
+def full_like(
+    other: Dataset | DataArray | Variable,
+    fill_value: Any,
+    dtype: DTypeMaybeMapping = None,
+) -> Dataset | DataArray | Variable:
+    ...
+
+
+def full_like(
+    other: Dataset | DataArray | Variable,
+    fill_value: Any,
+    dtype: DTypeMaybeMapping = None,
+) -> Dataset | DataArray | Variable:
     """Return a new object with the same shape and type as a given object.
 
     Parameters
@@ -1648,26 +1488,26 @@ def full_like(other, fill_value, dtype=None):
             f"fill_value must be scalar or, for datasets, a dict-like. Received {fill_value} instead."
         )
 
-    if not isinstance(other, Dataset) and isinstance(dtype, Mapping):
-        raise ValueError(
-            "'dtype' cannot be dict-like when passing a DataArray or Variable"
-        )
-
     if isinstance(other, Dataset):
         if not isinstance(fill_value, dict):
             fill_value = {k: fill_value for k in other.data_vars.keys()}
 
+        dtype_: Mapping[Any, DTypeLikeSave]
         if not isinstance(dtype, Mapping):
             dtype_ = {k: dtype for k in other.data_vars.keys()}
         else:
             dtype_ = dtype
 
         data_vars = {
-            k: _full_like_variable(v, fill_value.get(k, dtypes.NA), dtype_.get(k, None))
+            k: _full_like_variable(
+                v.variable, fill_value.get(k, dtypes.NA), dtype_.get(k, None)
+            )
             for k, v in other.data_vars.items()
         }
         return Dataset(data_vars, coords=other.coords, attrs=other.attrs)
     elif isinstance(other, DataArray):
+        if isinstance(dtype, Mapping):
+            raise ValueError("'dtype' cannot be dict-like when passing a DataArray")
         return DataArray(
             _full_like_variable(other.variable, fill_value, dtype),
             dims=other.dims,
@@ -1676,12 +1516,16 @@ def full_like(other, fill_value, dtype=None):
             name=other.name,
         )
     elif isinstance(other, Variable):
+        if isinstance(dtype, Mapping):
+            raise ValueError("'dtype' cannot be dict-like when passing a Variable")
         return _full_like_variable(other, fill_value, dtype)
     else:
         raise TypeError("Expected DataArray, Dataset, or Variable")
 
 
-def _full_like_variable(other, fill_value, dtype: DTypeLike = None):
+def _full_like_variable(
+    other: Variable, fill_value: Any, dtype: DTypeLike = None
+) -> Variable:
     """Inner function of full_like, where other must be a variable"""
     from .variable import Variable
 
@@ -1702,7 +1546,38 @@ def _full_like_variable(other, fill_value, dtype: DTypeLike = None):
     return Variable(dims=other.dims, data=data, attrs=other.attrs)
 
 
-def zeros_like(other, dtype: DTypeLike = None):
+@overload
+def zeros_like(other: DataArray, dtype: DTypeLikeSave = None) -> DataArray:
+    ...
+
+
+@overload
+def zeros_like(other: Dataset, dtype: DTypeMaybeMapping = None) -> Dataset:
+    ...
+
+
+@overload
+def zeros_like(other: Variable, dtype: DTypeLikeSave = None) -> Variable:
+    ...
+
+
+@overload
+def zeros_like(
+    other: Dataset | DataArray, dtype: DTypeMaybeMapping = None
+) -> Dataset | DataArray:
+    ...
+
+
+@overload
+def zeros_like(
+    other: Dataset | DataArray | Variable, dtype: DTypeMaybeMapping = None
+) -> Dataset | DataArray | Variable:
+    ...
+
+
+def zeros_like(
+    other: Dataset | DataArray | Variable, dtype: DTypeMaybeMapping = None
+) -> Dataset | DataArray | Variable:
     """Return a new object of zeros with the same shape and
     type as a given dataarray or dataset.
 
@@ -1758,7 +1633,38 @@ def zeros_like(other, dtype: DTypeLike = None):
     return full_like(other, 0, dtype)
 
 
-def ones_like(other, dtype: DTypeLike = None):
+@overload
+def ones_like(other: DataArray, dtype: DTypeLikeSave = None) -> DataArray:
+    ...
+
+
+@overload
+def ones_like(other: Dataset, dtype: DTypeMaybeMapping = None) -> Dataset:
+    ...
+
+
+@overload
+def ones_like(other: Variable, dtype: DTypeLikeSave = None) -> Variable:
+    ...
+
+
+@overload
+def ones_like(
+    other: Dataset | DataArray, dtype: DTypeMaybeMapping = None
+) -> Dataset | DataArray:
+    ...
+
+
+@overload
+def ones_like(
+    other: Dataset | DataArray | Variable, dtype: DTypeMaybeMapping = None
+) -> Dataset | DataArray | Variable:
+    ...
+
+
+def ones_like(
+    other: Dataset | DataArray | Variable, dtype: DTypeMaybeMapping = None
+) -> Dataset | DataArray | Variable:
     """Return a new object of ones with the same shape and
     type as a given dataarray or dataset.
 
@@ -1812,7 +1718,7 @@ def get_chunksizes(
 
     chunks: dict[Any, tuple[int, ...]] = {}
     for v in variables:
-        if hasattr(v.data, "chunks"):
+        if hasattr(v._data, "chunks"):
             for dim, c in v.chunksizes.items():
                 if dim in chunks and c != chunks[dim]:
                     raise ValueError(
@@ -1851,7 +1757,10 @@ def _contains_cftime_datetimes(array) -> bool:
 
 def contains_cftime_datetimes(var) -> bool:
     """Check if an xarray.Variable contains cftime.datetime objects"""
-    return _contains_cftime_datetimes(var.data)
+    if var.dtype == np.dtype("O") and var.size > 0:
+        return _contains_cftime_datetimes(var.data)
+    else:
+        return False
 
 
 def _contains_datetime_like_objects(var) -> bool:
