@@ -1,10 +1,14 @@
 """String formatting routines for __repr__.
 """
+from __future__ import annotations
+
 import contextlib
 import functools
+import math
+from collections import defaultdict
 from datetime import datetime, timedelta
 from itertools import chain, zip_longest
-from typing import Hashable
+from typing import Collection, Hashable
 
 import numpy as np
 import pandas as pd
@@ -41,10 +45,10 @@ def wrap_indent(text, start="", length=None):
 
 
 def _get_indexer_at_least_n_items(shape, n_desired, from_end):
-    assert 0 < n_desired <= np.prod(shape)
+    assert 0 < n_desired <= math.prod(shape)
     cum_items = np.cumprod(shape[::-1])
     n_steps = np.argmax(cum_items >= n_desired)
-    stop = int(np.ceil(float(n_desired) / np.r_[1, cum_items][n_steps]))
+    stop = math.ceil(float(n_desired) / np.r_[1, cum_items][n_steps])
     indexer = (
         ((-1 if from_end else 0),) * (len(shape) - 1 - n_steps)
         + ((slice(-stop, None) if from_end else slice(stop)),)
@@ -97,6 +101,16 @@ def last_item(array):
     return np.ravel(np.asarray(array[indexer])).tolist()
 
 
+def calc_max_rows_first(max_rows: int) -> int:
+    """Calculate the first rows to maintain the max number of rows."""
+    return max_rows // 2 + max_rows % 2
+
+
+def calc_max_rows_last(max_rows: int) -> int:
+    """Calculate the last rows to maintain the max number of rows."""
+    return max_rows // 2
+
+
 def format_timestamp(t):
     """Cast given object to a Timestamp and return a nicely formatted string"""
     # Timestamp is only valid for 1678 to 2262
@@ -143,7 +157,7 @@ def format_item(x, timedelta_format=None, quote_strings=True):
     elif isinstance(x, (str, bytes)):
         return repr(x) if quote_strings else x
     elif hasattr(x, "dtype") and np.issubdtype(x.dtype, np.floating):
-        return f"{x:.4}"
+        return f"{x.item():.4}"
     else:
         return str(x)
 
@@ -172,9 +186,7 @@ def format_array_flat(array, max_width: int):
     """
     # every item will take up at least two characters, but we always want to
     # print at least first and last items
-    max_possibly_relevant = min(
-        max(array.size, 1), max(int(np.ceil(max_width / 2.0)), 2)
-    )
+    max_possibly_relevant = min(max(array.size, 1), max(math.ceil(max_width / 2.0), 2))
     relevant_front_items = format_items(
         first_n_items(array, (max_possibly_relevant + 1) // 2)
     )
@@ -190,7 +202,7 @@ def format_array_flat(array, max_width: int):
         (max_possibly_relevant < array.size) or (cum_len > max_width).any()
     ):
         padding = " ... "
-        max_len = max(int(np.argmax(cum_len + len(padding) - 1 > max_width)), 2)  # type: ignore[type-var]
+        max_len = max(int(np.argmax(cum_len + len(padding) - 1 > max_width)), 2)
         count = min(array.size, max_len)
     else:
         count = array.size
@@ -256,10 +268,10 @@ def inline_sparse_repr(array):
 
 def inline_variable_array_repr(var, max_width):
     """Build a one-line summary of a variable's data."""
-    if var._in_memory:
-        return format_array_flat(var, max_width)
-    elif hasattr(var._data, "_repr_inline_"):
+    if hasattr(var._data, "_repr_inline_"):
         return var._data._repr_inline_(max_width)
+    elif var._in_memory:
+        return format_array_flat(var, max_width)
     elif isinstance(var._data, dask_array_type):
         return inline_dask_repr(var.data)
     elif isinstance(var._data, sparse_array_type):
@@ -272,68 +284,33 @@ def inline_variable_array_repr(var, max_width):
 
 
 def summarize_variable(
-    name: Hashable, var, col_width: int, marker: str = " ", max_width: int = None
+    name: Hashable, var, col_width: int, max_width: int = None, is_index: bool = False
 ):
     """Summarize a variable in one line, e.g., for the Dataset.__repr__."""
+    variable = getattr(var, "variable", var)
+
     if max_width is None:
         max_width_options = OPTIONS["display_width"]
         if not isinstance(max_width_options, int):
             raise TypeError(f"`max_width` value of `{max_width}` is not a valid int")
         else:
             max_width = max_width_options
+
+    marker = "*" if is_index else " "
     first_col = pretty_print(f"  {marker} {name} ", col_width)
-    if var.dims:
-        dims_str = "({}) ".format(", ".join(map(str, var.dims)))
+
+    if variable.dims:
+        dims_str = "({}) ".format(", ".join(map(str, variable.dims)))
     else:
         dims_str = ""
-    front_str = f"{first_col}{dims_str}{var.dtype} "
+    front_str = f"{first_col}{dims_str}{variable.dtype} "
 
     values_width = max_width - len(front_str)
-    values_str = inline_variable_array_repr(var, values_width)
+    values_str = inline_variable_array_repr(variable, values_width)
 
     return front_str + values_str
 
 
-def _summarize_coord_multiindex(coord, col_width, marker):
-    first_col = pretty_print(f"  {marker} {coord.name} ", col_width)
-    return "{}({}) MultiIndex".format(first_col, str(coord.dims[0]))
-
-
-def _summarize_coord_levels(coord, col_width, marker="-"):
-    if len(coord) > 100 and col_width < len(coord):
-        n_values = col_width
-        indices = list(range(0, n_values)) + list(range(-n_values, 0))
-        subset = coord[indices]
-    else:
-        subset = coord
-
-    return "\n".join(
-        summarize_variable(
-            lname, subset.get_level_variable(lname), col_width, marker=marker
-        )
-        for lname in subset.level_names
-    )
-
-
-def summarize_datavar(name, var, col_width):
-    return summarize_variable(name, var.variable, col_width)
-
-
-def summarize_coord(name: Hashable, var, col_width: int):
-    is_index = name in var.dims
-    marker = "*" if is_index else " "
-    if is_index:
-        coord = var.variable.to_index_variable()
-        if coord.level_names is not None:
-            return "\n".join(
-                [
-                    _summarize_coord_multiindex(coord, col_width, marker),
-                    _summarize_coord_levels(coord, col_width),
-                ]
-            )
-    return summarize_variable(name, var.variable, col_width, marker)
-
-
 def summarize_attr(key, value, col_width=None):
     """Summary for __repr__ - use ``X.attrs[key]`` for full value."""
     # Indent key and add ':', then right-pad if col_width is not None
@@ -349,23 +326,6 @@ def summarize_attr(key, value, col_width=None):
 EMPTY_REPR = "    *empty*"
 
 
-def _get_col_items(mapping):
-    """Get all column items to format, including both keys of `mapping`
-    and MultiIndex levels if any.
-    """
-    from .variable import IndexVariable
-
-    col_items = []
-    for k, v in mapping.items():
-        col_items.append(k)
-        var = getattr(v, "variable", v)
-        if isinstance(var, IndexVariable):
-            level_names = var.to_index_variable().level_names
-            if level_names is not None:
-                col_items += list(level_names)
-    return col_items
-
-
 def _calculate_col_width(col_items):
     max_name_length = max(len(str(s)) for s in col_items) if col_items else 0
     col_width = max(max_name_length, 7) + 6
@@ -373,10 +333,21 @@ def _calculate_col_width(col_items):
 
 
 def _mapping_repr(
-    mapping, title, summarizer, expand_option_name, col_width=None, max_rows=None
+    mapping,
+    title,
+    summarizer,
+    expand_option_name,
+    col_width=None,
+    max_rows=None,
+    indexes=None,
 ):
     if col_width is None:
         col_width = _calculate_col_width(mapping)
+
+    summarizer_kwargs = defaultdict(dict)
+    if indexes is not None:
+        summarizer_kwargs = {k: {"is_index": k in indexes} for k in mapping}
+
     summary = [f"{title}:"]
     if mapping:
         len_mapping = len(mapping)
@@ -384,17 +355,24 @@ def _mapping_repr(
             summary = [f"{summary[0]} ({len_mapping})"]
         elif max_rows is not None and len_mapping > max_rows:
             summary = [f"{summary[0]} ({max_rows}/{len_mapping})"]
-            first_rows = max_rows // 2 + max_rows % 2
+            first_rows = calc_max_rows_first(max_rows)
             keys = list(mapping.keys())
-            summary += [summarizer(k, mapping[k], col_width) for k in keys[:first_rows]]
+            summary += [
+                summarizer(k, mapping[k], col_width, **summarizer_kwargs[k])
+                for k in keys[:first_rows]
+            ]
             if max_rows > 1:
-                last_rows = max_rows // 2
+                last_rows = calc_max_rows_last(max_rows)
                 summary += [pretty_print("    ...", col_width) + " ..."]
                 summary += [
-                    summarizer(k, mapping[k], col_width) for k in keys[-last_rows:]
+                    summarizer(k, mapping[k], col_width, **summarizer_kwargs[k])
+                    for k in keys[-last_rows:]
                 ]
         else:
-            summary += [summarizer(k, v, col_width) for k, v in mapping.items()]
+            summary += [
+                summarizer(k, v, col_width, **summarizer_kwargs[k])
+                for k, v in mapping.items()
+            ]
     else:
         summary += [EMPTY_REPR]
     return "\n".join(summary)
@@ -403,7 +381,7 @@ def _mapping_repr(
 data_vars_repr = functools.partial(
     _mapping_repr,
     title="Data variables",
-    summarizer=summarize_datavar,
+    summarizer=summarize_variable,
     expand_option_name="display_expand_data_vars",
 )
 
@@ -418,21 +396,25 @@ attrs_repr = functools.partial(
 
 def coords_repr(coords, col_width=None, max_rows=None):
     if col_width is None:
-        col_width = _calculate_col_width(_get_col_items(coords))
+        col_width = _calculate_col_width(coords)
     return _mapping_repr(
         coords,
         title="Coordinates",
-        summarizer=summarize_coord,
+        summarizer=summarize_variable,
         expand_option_name="display_expand_coords",
         col_width=col_width,
+        indexes=coords.xindexes,
         max_rows=max_rows,
     )
 
 
 def indexes_repr(indexes):
-    summary = []
-    for k, v in indexes.items():
-        summary.append(wrap_indent(repr(v), f"{k}: "))
+    summary = ["Indexes:"]
+    if indexes:
+        for k, v in indexes.items():
+            summary.append(wrap_indent(repr(v), f"{k}: "))
+    else:
+        summary += [EMPTY_REPR]
     return "\n".join(summary)
 
 
@@ -441,11 +423,74 @@ def dim_summary(obj):
     return ", ".join(elements)
 
 
-def unindexed_dims_repr(dims, coords):
+def _element_formatter(
+    elements: Collection[Hashable],
+    col_width: int,
+    max_rows: int | None = None,
+    delimiter: str = ", ",
+) -> str:
+    """
+    Formats elements for better readability.
+
+    Once it becomes wider than the display width it will create a newline and
+    continue indented to col_width.
+    Once there are more rows than the maximum displayed rows it will start
+    removing rows.
+
+    Parameters
+    ----------
+    elements : Collection of hashable
+        Elements to join together.
+    col_width : int
+        The width to indent to if a newline has been made.
+    max_rows : int, optional
+        The maximum number of allowed rows. The default is None.
+    delimiter : str, optional
+        Delimiter to use between each element. The default is ", ".
+    """
+    elements_len = len(elements)
+    out = [""]
+    length_row = 0
+    for i, v in enumerate(elements):
+        delim = delimiter if i < elements_len - 1 else ""
+        v_delim = f"{v}{delim}"
+        length_element = len(v_delim)
+        length_row += length_element
+
+        # Create a new row if the next elements makes the print wider than
+        # the maximum display width:
+        if col_width + length_row > OPTIONS["display_width"]:
+            out[-1] = out[-1].rstrip()  # Remove trailing whitespace.
+            out.append("\n" + pretty_print("", col_width) + v_delim)
+            length_row = length_element
+        else:
+            out[-1] += v_delim
+
+    # If there are too many rows of dimensions trim some away:
+    if max_rows and (len(out) > max_rows):
+        first_rows = calc_max_rows_first(max_rows)
+        last_rows = calc_max_rows_last(max_rows)
+        out = (
+            out[:first_rows]
+            + ["\n" + pretty_print("", col_width) + "..."]
+            + (out[-last_rows:] if max_rows > 1 else [])
+        )
+    return "".join(out)
+
+
+def dim_summary_limited(obj, col_width: int, max_rows: int | None = None) -> str:
+    elements = [f"{k}: {v}" for k, v in obj.sizes.items()]
+    return _element_formatter(elements, col_width, max_rows)
+
+
+def unindexed_dims_repr(dims, coords, max_rows: int | None = None):
     unindexed_dims = [d for d in dims if d not in coords]
     if unindexed_dims:
-        dims_str = ", ".join(f"{d}" for d in unindexed_dims)
-        return "Dimensions without coordinates: " + dims_str
+        dims_start = "Dimensions without coordinates: "
+        dims_str = _element_formatter(
+            unindexed_dims, col_width=len(dims_start), max_rows=max_rows
+        )
+        return dims_start + dims_str
     else:
         return None
 
@@ -476,7 +521,11 @@ def short_numpy_repr(array):
 
     # default to lower precision so a full (abbreviated) line can fit on
     # one line with the default display_width
-    options = {"precision": 6, "linewidth": OPTIONS["display_width"], "threshold": 200}
+    options = {
+        "precision": 6,
+        "linewidth": OPTIONS["display_width"],
+        "threshold": OPTIONS["display_values_threshold"],
+    }
     if array.ndim < 3:
         edgeitems = 3
     elif array.ndim == 3:
@@ -505,6 +554,8 @@ def short_data_repr(array):
 def array_repr(arr):
     from .variable import Variable
 
+    max_rows = OPTIONS["display_max_rows"]
+
     # used for DataArray, Variable and IndexVariable
     if hasattr(arr, "name") and arr.name is not None:
         name_str = f"{arr.name!r} "
@@ -520,16 +571,23 @@ def array_repr(arr):
     else:
         data_repr = inline_variable_array_repr(arr.variable, OPTIONS["display_width"])
 
+    start = f"<xarray.{type(arr).__name__} {name_str}"
+    dims = dim_summary_limited(arr, col_width=len(start) + 1, max_rows=max_rows)
     summary = [
-        "<xarray.{} {}({})>".format(type(arr).__name__, name_str, dim_summary(arr)),
+        f"{start}({dims})>",
         data_repr,
     ]
 
     if hasattr(arr, "coords"):
         if arr.coords:
-            summary.append(repr(arr.coords))
+            col_width = _calculate_col_width(arr.coords)
+            summary.append(
+                coords_repr(arr.coords, col_width=col_width, max_rows=max_rows)
+            )
 
-        unindexed_dims_str = unindexed_dims_repr(arr.dims, arr.coords)
+        unindexed_dims_str = unindexed_dims_repr(
+            arr.dims, arr.coords, max_rows=max_rows
+        )
         if unindexed_dims_str:
             summary.append(unindexed_dims_str)
 
@@ -540,18 +598,19 @@ def array_repr(arr):
 
 
 def dataset_repr(ds):
-    summary = ["<xarray.{}>".format(type(ds).__name__)]
+    summary = [f"<xarray.{type(ds).__name__}>"]
 
-    col_width = _calculate_col_width(_get_col_items(ds.variables))
+    col_width = _calculate_col_width(ds.variables)
     max_rows = OPTIONS["display_max_rows"]
 
     dims_start = pretty_print("Dimensions:", col_width)
-    summary.append("{}({})".format(dims_start, dim_summary(ds)))
+    dims_values = dim_summary_limited(ds, col_width=col_width + 1, max_rows=max_rows)
+    summary.append(f"{dims_start}({dims_values})")
 
     if ds.coords:
         summary.append(coords_repr(ds.coords, col_width=col_width, max_rows=max_rows))
 
-    unindexed_dims_str = unindexed_dims_repr(ds.dims, ds.coords)
+    unindexed_dims_str = unindexed_dims_repr(ds.dims, ds.coords, max_rows=max_rows)
     if unindexed_dims_str:
         summary.append(unindexed_dims_str)
 
@@ -572,9 +631,20 @@ def diff_dim_summary(a, b):
         return ""
 
 
-def _diff_mapping_repr(a_mapping, b_mapping, compat, title, summarizer, col_width=None):
-    def extra_items_repr(extra_keys, mapping, ab_side):
-        extra_repr = [summarizer(k, mapping[k], col_width) for k in extra_keys]
+def _diff_mapping_repr(
+    a_mapping,
+    b_mapping,
+    compat,
+    title,
+    summarizer,
+    col_width=None,
+    a_indexes=None,
+    b_indexes=None,
+):
+    def extra_items_repr(extra_keys, mapping, ab_side, kwargs):
+        extra_repr = [
+            summarizer(k, mapping[k], col_width, **kwargs[k]) for k in extra_keys
+        ]
         if extra_repr:
             header = f"{title} only on the {ab_side} object:"
             return [header] + extra_repr
@@ -588,6 +658,13 @@ def _diff_mapping_repr(a_mapping, b_mapping, compat, title, summarizer, col_widt
 
     diff_items = []
 
+    a_summarizer_kwargs = defaultdict(dict)
+    if a_indexes is not None:
+        a_summarizer_kwargs = {k: {"is_index": k in a_indexes} for k in a_mapping}
+    b_summarizer_kwargs = defaultdict(dict)
+    if b_indexes is not None:
+        b_summarizer_kwargs = {k: {"is_index": k in b_indexes} for k in b_mapping}
+
     for k in a_keys & b_keys:
         try:
             # compare xarray variable
@@ -607,7 +684,8 @@ def _diff_mapping_repr(a_mapping, b_mapping, compat, title, summarizer, col_widt
 
         if not compatible:
             temp = [
-                summarizer(k, vars[k], col_width) for vars in (a_mapping, b_mapping)
+                summarizer(k, a_mapping[k], col_width, **a_summarizer_kwargs[k]),
+                summarizer(k, b_mapping[k], col_width, **b_summarizer_kwargs[k]),
             ]
 
             if compat == "identical" and is_variable:
@@ -629,19 +707,29 @@ def _diff_mapping_repr(a_mapping, b_mapping, compat, title, summarizer, col_widt
     if diff_items:
         summary += [f"Differing {title.lower()}:"] + diff_items
 
-    summary += extra_items_repr(a_keys - b_keys, a_mapping, "left")
-    summary += extra_items_repr(b_keys - a_keys, b_mapping, "right")
+    summary += extra_items_repr(a_keys - b_keys, a_mapping, "left", a_summarizer_kwargs)
+    summary += extra_items_repr(
+        b_keys - a_keys, b_mapping, "right", b_summarizer_kwargs
+    )
 
     return "\n".join(summary)
 
 
-diff_coords_repr = functools.partial(
-    _diff_mapping_repr, title="Coordinates", summarizer=summarize_coord
-)
+def diff_coords_repr(a, b, compat, col_width=None):
+    return _diff_mapping_repr(
+        a,
+        b,
+        compat,
+        "Coordinates",
+        summarize_variable,
+        col_width=col_width,
+        a_indexes=a.indexes,
+        b_indexes=b.indexes,
+    )
 
 
 diff_data_vars_repr = functools.partial(
-    _diff_mapping_repr, title="Data variables", summarizer=summarize_datavar
+    _diff_mapping_repr, title="Data variables", summarizer=summarize_variable
 )
 
 
@@ -703,9 +791,7 @@ def diff_dataset_repr(a, b, compat):
         )
     ]
 
-    col_width = _calculate_col_width(
-        set(_get_col_items(a.variables) + _get_col_items(b.variables))
-    )
+    col_width = _calculate_col_width(set(list(a.variables) + list(b.variables)))
 
     summary.append(diff_dim_summary(a, b))
     summary.append(diff_coords_repr(a.coords, b.coords, compat, col_width=col_width))
