@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import functools
 import itertools
 import warnings
@@ -9,8 +11,8 @@ from .utils import (
     _get_nice_quiver_magnitude,
     _infer_xy_labels,
     _process_cmap_cbar_kwargs,
+    import_matplotlib_pyplot,
     label_from_attrs,
-    plt,
 )
 
 # Overrides axes.labelsize, xtick.major.size, ytick.major.size
@@ -93,7 +95,7 @@ class FacetGrid:
         data : DataArray
             xarray DataArray to be plotted.
         row, col : str
-            Dimesion names that define subsets of the data, which will be drawn
+            Dimension names that define subsets of the data, which will be drawn
             on separate facets in the grid.
         col_wrap : int, optional
             "Wrap" the grid the for the column variable after this number of columns,
@@ -116,6 +118,8 @@ class FacetGrid:
 
         """
 
+        plt = import_matplotlib_pyplot()
+
         # Handle corner case of nonunique coordinates
         rep_col = col is not None and not data[col].to_index().is_unique
         rep_row = row is not None and not data[row].to_index().is_unique
@@ -517,8 +521,10 @@ class FacetGrid:
         self: FacetGrid object
 
         """
+        import matplotlib as mpl
+
         if size is None:
-            size = plt.rcParams["axes.labelsize"]
+            size = mpl.rcParams["axes.labelsize"]
 
         nicetitle = functools.partial(_nicetitle, maxchar=maxchar, template=template)
 
@@ -615,6 +621,8 @@ class FacetGrid:
         self : FacetGrid object
 
         """
+        plt = import_matplotlib_pyplot()
+
         for ax, namedict in zip(self.axes.flat, self.name_dicts.flat):
             if namedict is not None:
                 data = self.data.loc[namedict]
