@@ -190,7 +190,7 @@ STUBFILE_PREAMBLE = '''\
 """Stub file for mixin classes with arithmetic operators."""
 # This file was generated using xarray.util.generate_ops. Do not edit manually.
 
-from typing import NoReturn, TypeVar, Union, overload
+from typing import NoReturn, TypeVar, overload
 
 import numpy as np
 
@@ -210,7 +210,7 @@ from .variable import Variable
 try:
     from dask.array import Array as DaskArray
 except ImportError:
-    DaskArray = np.ndarray
+    DaskArray = np.ndarray  # type: ignore
 
 # DatasetOpsMixin etc. are parent classes of Dataset etc.
 # Because of https://github.com/pydata/xarray/issues/5755, we redefine these. Generally
